#include <termios.h> //tcgetattr
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <time.h>


#define ECFG_PROTO_VER
#include "pmc_e1_if.h"

#define ELF_ROUTER
#include "defs.h"
#include "menus.h"
#include "ecfg_ver.h"

#define DEV_TITLE "Quasar"
#define DEV_FILE "/dev/quasar"
#define DEV_CFG_FILE "/etc/ecfg/quasar"

#include "menus.c"
#include "ecfg_menus.c"

int fd = -1;
int board = 0;	  //E1 board(PMC) number
int chan_num = 0; //E1 channel number (0...7)
char log_file[256]="";
char conf_file[256]="";

FILE *log_stream = NULL;

#define LOG(b...) ecfg_log(log_stream,##b)
int ecfg_log (FILE *stream, char *fmt, ...) {
    va_list argptr;
    int cnt;
    char timestr[256];
    time_t t;

    if (NULL == stream) return 0; //Nothing to do.

    t = time (NULL);
    ctime_r (&t, timestr);
    timestr[strlen(timestr)-1] = '\0';
    fprintf (stream, "%s: ", timestr);

    va_start (argptr, fmt);
    cnt = vfprintf (stream, fmt, argptr);
    va_end (argptr);

    fflush (stream);
    return cnt;
}

char getch_nolock (void) {
    char ch;
    int i;
    struct termios old_tc,new_tc;

    tcgetattr(fileno(stdin),&old_tc);
    new_tc = old_tc;
    new_tc.c_lflag &= ~ICANON;
    tcsetattr(fileno(stdin),TCSANOW,&new_tc);

    ch = i = getc(stdin);
    tcsetattr(fileno(stdin),TCSANOW,&old_tc);
    if (-1 == i) {
	LOG ("Terminal disconnected. Bye.\n");
        exit (1);
    }
    return ch;
}

void ecfg_manage_cfg (unsigned char dir) {
    int i;
    char buf[256]={0,}, tbuf[16];

    abcd.chan_num = cfg.chan_num = inserr.chan_num = chan_num;
    cfg.chan_num |= dir;
    abcd.chan_num |= dir;
//Mixel doesn't support read back !    inserr.chan_num |= dir;

    if (cfg.cfg0 & CFG0_RCV_HDB3) cfg.cfg0 |= CFG0_XMIT_HDB3;
    else cfg.cfg0 &= ~CFG0_XMIT_HDB3;

    LOG ("Pre-ioctl config\n");
    LOG ("cfg.chan_num=0x%x\n", cfg.chan_num);
//    LOG ("cfg.abcd=0x%x\n", cfg.abcd);
    LOG ("cfg.cfg0=0x%x\n", cfg.cfg0);
    LOG ("cfg.cfg1=0x%x\n", cfg.cfg1);
    LOG ("cfg.cfg2=0x%x\n", cfg.cfg2);
    LOG ("cfg.cfg3=0x%x\n", cfg.cfg3);
    LOG ("cfg.cfg4=0x%x\n", cfg.cfg4);
    i = ioctl (fd, PTMC_E1_CHAN_CONFIGURE, &cfg);
    if (i < 0) {
	perror ("ioctl(PTMC_E1_CHAN_CONFIGURE)");
	exit (-1);
    }
    LOG ("Post-ioctl config\n");
    LOG ("cfg.chan_num=0x%x\n", cfg.chan_num);
//    LOG ("cfg.abcd=0x%x\n", cfg.abcd);
    LOG ("cfg.cfg0=0x%x\n", cfg.cfg0);
    LOG ("cfg.cfg1=0x%x\n", cfg.cfg1);
    LOG ("cfg.cfg2=0x%x\n", cfg.cfg2);
    LOG ("cfg.cfg3=0x%x\n", cfg.cfg3);
    LOG ("cfg.cfg4=0x%x\n", cfg.cfg4);

    if (version.board_id == BOARD_QUASAR8) {
        LOG ("INSERR-ioctl\n");
        LOG ("inserr.chan_num=0x%x\n", inserr.chan_num);
        LOG ("inserr.pmc_err=0x%x\n", inserr.pmc_err);
        i = ioctl (fd, PTMC_E1_ERR_EMU, &inserr);
        if (i < 0) {
	       perror ("ioctl(PTMC_E1_ERR_EMU)");
	       exit (-1);
        }
        memset (&inserr, 0, sizeof (inserr));
    }

    LOG ("ABCD-ioctl\n");
    LOG ("abcd.chan_num=0x%x\n", abcd.chan_num);
    for (i=0; i<16; ++i) {
         sprintf (tbuf, "%x:", abcd.cdata[i]);
         strcat (buf, tbuf);
    }
    LOG ("%s\n", buf);
    i = ioctl (fd, PTMC_E1_CAS_XMIT_PASS, &abcd);
    if (i < 0) {
	perror ("ioctl(PTMC_E1_CAS_XMIT_PASS)");
	exit (-1);
    }
    ecfg_read_status ();
}

void ecfg_apply_cfg (void) {
    ecfg_manage_cfg (0);
}

void ecfg_read_status () {
    int i, reset_counters;

    reset_counters = status.mgmt;
    memset (&status, 0, sizeof (status));
    status.chan_num = chan_num;
    if (reset_counters) status.mgmt = 0x7;

    i = ioctl (fd, PTMC_E1_GET_STATS, &status);
    if (i < 0) {
	perror ("ioctl(PTMC_E1_GET_STATS)");
	exit (-1);
    }

    LOG ("Status\n");
    LOG ("status.chan_num=0x%x\n", status.chan_num);
    LOG ("status.stat0=0x%x\n", status.stat0);
    LOG ("status.stat1=0x%x\n", status.stat1);
    LOG ("status.stat2=0x%x\n", status.stat2);
    LOG ("status.mgmt=0x%x\n", status.mgmt);
    LOG ("status.ovr=0x%x\n", status.ovr);
    LOG ("status.hdb3=0x%x\n", status.hdb3);
    LOG ("status.crc4=0x%x\n", status.crc4);
    LOG ("status.fas=0x%x\n", status.fas);

    status.mgmt = 0;
}

void ecfg_factory_cfg (void) {
    int i;

    memset((void *) &cfg, 0, sizeof(cfg));
    if (version.board_id == BOARD_QUASAR8) {
        cfg.cfg0 = (CFG0_MASTER | CFG0_RCV_HDB3 | CFG0_XMIT_HDB3);
        cfg.cfg1 = 0;
        cfg.cfg2 = CFG2_X1 | CFG2_X2 | CFG2_X3;
        cfg.cfg3 = CFG3_CRC4_MANUAL;
        cfg.cfg4 = CFG4_AIS;
    }
    else {
        cfg.cfg0 = (CFG0_MASTER | CFG0_RCV_HDB3 | CFG0_XMIT_HDB3);
        cfg.cfg1 = 0;
        cfg.cfg2 = (CFG2_RAI_MANUAL | CFG2_RCAS_MANUAL \
             | CFG2_X1 | CFG2_X2 | CFG2_X3);
        cfg.cfg3 = CFG3_CRC4_MANUAL;
        cfg.cfg4 = CFG4_AIS;
    }
    for (i=0; i<16; ++i) {
        abcd.cdata[i] = 0xAA;
    }
    memset (&inserr, 0, sizeof (inserr));
}

void ecfg_save_cfg (void) {
    FILE *stream;

    stream = fopen (conf_file, "w");
    if (NULL == stream) {
	fprintf (stderr, "Failed to open configuration file \"%s\": ",
		 conf_file);
	perror ("");
	exit (1);
    }
    if (fwrite (&cfg, sizeof (cfg), 1, stream) != 1) {
	fprintf (stderr, "Failed to save configuration \"%s\": ", conf_file);
	perror ("");
	fclose (stream);
        exit (1);
    }
    if (fwrite (&abcd, sizeof (abcd), 1, stream) != 1) {
	fprintf (stderr, "Failed to save configuration \"%s\": ", conf_file);
	perror ("");
	fclose (stream);
        exit (1);
    }
    fclose (stream);
}

void ecfg_restore_cfg (void) {
    FILE *stream;

    stream = fopen (conf_file, "r");
    if (NULL == stream) {
	fprintf (stderr, "Failed to open configuration file \"%s\": ",
		 conf_file);
	perror ("");
	exit (1);
    }
    if (fread (&cfg, sizeof (cfg), 1, stream) != 1) {
	fprintf (stderr, "Failed to restore configuration \"%s\": ", conf_file);
	perror ("");
        fclose (stream);
	exit (1);
    }
    if (fread (&abcd, sizeof (abcd), 1, stream) != 1) {
	fprintf (stderr, "Failed to restore configuration \"%s\": ", conf_file);
	perror ("");
        fclose (stream);
	exit (1);
    }
    fclose (stream);
}


void construct_cfg_name (void) {
    if ('\0' == conf_file[0]) {
	snprintf (conf_file, sizeof (conf_file),
		  DEV_CFG_FILE "%d_%d.cfg", board, chan_num);
    }
}

void ecfg_read_version () {
    int i, reset_counters;

    reset_counters = status.mgmt;
    memset (&status, 0, sizeof (status));
    status.chan_num = chan_num;
    if (reset_counters) status.mgmt = 0x7;

    i = ioctl (fd, PTMC_E1_GET_VERSION, &version);
    if (i < 0) {
	perror ("ioctl(PTMC_E1_GET_VERSION)");
	exit (-1);
    }

    LOG ("Version\n");
    LOG ("version.proto_version=%s, ecfg_proto_version=%s\n",
	 version.proto_version, ecfg_proto_version);
    if (strncmp (version.proto_version, ecfg_proto_version,
                  sizeof (version.proto_version))) {
        fprintf (stderr, "Protocol version mismatch (driver: %s, ecfg: %s).\n"
                "Please, use ecfg shipped with the driver.\n",
                version.proto_version, ecfg_proto_version);
        exit (-1);
    }
    LOG ("version.hw_fw_rev=0x%x\n", version.hw_fw_rev);
    LOG ("version.driver_version=0x%x\n", version.driver_version);
}

void help_and_exit (const char *prog) {
    construct_cfg_name ();
    ecfg_print_title ();
    printf ("%s [options]\n"
	    "\t-f {file} - Configuration file path\n"
	    "\t-F - Load factory settings\n"
	    "\t-l - Load (run) configuration from file (no menu)\n"
	    "\t-b {value} - E1 Board number\n"
	    "\t-i {value} - E1 channel number\n"
	    "\t-D {file} - Dump debug information to file\n"
	    "\t-h - Help\n", prog);
    exit (1);
}

#define OPS_LOADCFG 0x1
#define OPS_FACTORY 0x2
//#define DEBUG_MENU 0x2

int main (int argc, char **argv) {
    int opt;
    char bus[256];
    unsigned char flags = 0;
    chan_num = 0;
    bus[0] = '\0';

    ecfg_factory_cfg ();

    while ((opt = getopt(argc, argv, "f:Flb:i:D:h")) != -1) {
	switch (opt) {
	case 'f':
	    strncpy (conf_file, optarg, sizeof (conf_file));
	    break;
	case 'F':
            flags |= OPS_FACTORY;
            break;
	case 'l':
	    flags |= OPS_LOADCFG;
	    break;
	case 'b':
	    board = atoi (optarg);
	    break;
	case 'i':
	    chan_num = atoi (optarg);
	    break;
	case 'D':
	    strncpy (log_file, optarg, sizeof (log_file));
	    break;
	case 'h':
	default:
	    help_and_exit (argv[0]);
	}
    }

    if ('\0' != log_file[0]) {
	log_stream = fopen (log_file, "w");
	if (NULL == log_stream) {
	    fprintf (stderr, "Failed to open log file \"%s\": ", log_file);
	    perror ("");
	    exit (1);
	}
    }

    construct_cfg_name ();

    sprintf (bus, DEV_FILE "%d", board);

    LOG ("Using %d for chan %d\n\n", board, chan_num);

    fd = open (bus, O_RDONLY);
    if (fd < 0) {
	printf ("Can't open %s\n", bus);
	exit (-1);
    }

    ecfg_read_version ();
    ecfg_read_status ();

    if (flags & OPS_LOADCFG) {
	printf ("Loading configuration from %s ... ", conf_file);
	ecfg_restore_cfg ();
	ecfg_apply_cfg ();
	printf ("Done.\n");
	exit (0);
    }

    if (!(flags & OPS_FACTORY)) {
	ecfg_manage_cfg (DIRECT_IN);
    }
    ecfg_apply_cfg ();
    menu_init ();
    menu ();
    return 0;
}
